"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const t = __importStar(require("@babel/types"));
const graphql_tool_utilities_1 = require("graphql-tool-utilities");
const types_1 = require("../../types");
const context_1 = require("./context");
const utilities_1 = require("./utilities");
const language_1 = require("./language");
const generate = require('@babel/generator').default;
function printDocument({ path, operation, fragments }, ast, options) {
    const file = new context_1.FileContext(path, options);
    if (operation == null) {
        const fileBody = fragments.reduce((statements, fragment) => {
            const context = new context_1.OperationContext(fragment, ast, options, file);
            const body = language_1.tsInterfaceBodyForObjectField(fragment, fragment.typeCondition, new utilities_1.ObjectStack(fragment.typeCondition), context);
            const { namespace } = context;
            return [
                ...statements,
                ...(namespace ? [t.exportNamedDeclaration(namespace, [])] : []),
                t.exportNamedDeclaration(t.tsInterfaceDeclaration(t.identifier(context.typeName), null, null, body), []),
            ];
        }, []);
        const { schemaImports } = file;
        if (schemaImports) {
            fileBody.unshift(schemaImports);
        }
        return generate(t.file(t.program(fileBody), [], [])).code;
    }
    const context = new context_1.OperationContext(operation, ast, options, file);
    const partialContext = new context_1.OperationContext(operation, ast, Object.assign(Object.assign({}, options), { partial: true }), file);
    let rootType;
    if (operation.operationType === graphql_tool_utilities_1.OperationType.Query) {
        rootType = ast.schema.getQueryType();
    }
    else if (operation.operationType === graphql_tool_utilities_1.OperationType.Mutation) {
        rootType = ast.schema.getMutationType();
    }
    else {
        rootType = ast.schema.getSubscriptionType();
    }
    const variables = operation.variables.filter(graphql_tool_utilities_1.isTypedVariable).length > 0
        ? context.export(language_1.variablesInterface(operation.variables, context))
        : null;
    const operationInterface = t.tsInterfaceDeclaration(t.identifier(context.typeName), null, null, language_1.tsInterfaceBodyForObjectField(operation, rootType, new utilities_1.ObjectStack(rootType), context));
    const operationPartialInterface = t.tsInterfaceDeclaration(t.identifier(partialContext.typeName), null, null, language_1.tsInterfaceBodyForObjectField(operation, rootType, new utilities_1.ObjectStack(rootType), partialContext));
    const { schemaImports } = file;
    const { namespace } = context;
    const { namespace: partialNamespace } = partialContext;
    const { exportFormat = types_1.ExportFormat.Document } = options;
    const documentType = exportFormat === types_1.ExportFormat.Document ? 'DocumentNode' : 'SimpleDocument';
    const documentNodeImport = t.importDeclaration([t.importSpecifier(t.identifier(documentType), t.identifier(documentType))], t.stringLiteral('graphql-typed'));
    const documentNodeDeclaratorIdentifier = t.identifier('document');
    documentNodeDeclaratorIdentifier.typeAnnotation = t.tsTypeAnnotation(t.tsTypeReference(t.identifier(documentType), t.tsTypeParameterInstantiation([
        t.tsTypeReference(t.identifier(context.typeName)),
        variables || t.tsNeverKeyword(),
        t.tsTypeReference(t.identifier(partialContext.typeName)),
    ])));
    const documentNodeDeclaration = t.variableDeclaration('const', [
        t.variableDeclarator(documentNodeDeclaratorIdentifier),
    ]);
    documentNodeDeclaration.declare = true;
    const documentNodeExport = t.exportDefaultDeclaration(t.identifier('document'));
    const fileBody = [documentNodeImport];
    if (schemaImports) {
        fileBody.push(schemaImports);
    }
    if (partialNamespace) {
        fileBody.push(t.exportNamedDeclaration(partialNamespace, []));
    }
    fileBody.push(t.exportNamedDeclaration(operationPartialInterface, []));
    if (namespace) {
        fileBody.push(t.exportNamedDeclaration(namespace, []));
    }
    fileBody.push(t.exportNamedDeclaration(operationInterface, []), documentNodeDeclaration, documentNodeExport);
    return generate(t.file(t.program(fileBody), [], [])).code;
}
exports.printDocument = printDocument;
